# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.7
#   kernelspec:
#     display_name: gpjax
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Likelihood guide
#
# In this notebook, we will walk users through the process of creating a new likelihood
# in GPJax.
#
# ## Background
#
# In this section we'll provide a short introduction to likelihoods and why they are
# important. For users who are already familiar with likelihoods, feel free to skip to
# the next section, and for users who would like more information than is provided
# here, please see our [introduction to Gaussian processes notebook](intro_to_gps.md).
#
# ### What is a likelihood?
#
# We adopt the notation of our
# [introduction to Gaussian processes notebook](intro_to_gps.md) where we have a
# Gaussian process (GP) $f(\cdot)\sim\mathcal{GP}(m(\cdot), k(\cdot, \cdot))$ and a
# dataset $\mathbf{y} = \{y_n\}_{n=1}^N$ observed at corresponding inputs
# $\mathbf{x} = \{x_n\}_{n=1}^N$. The evaluation of $f$ at $\mathbf{x}$ is denoted by
# $\mathbf{f} = \{f(x_n)\}_{n=1}^N$. The _likelihood function_ of the GP is then given
# by
# $$
# \begin{align}
#     \label{eq:likelihood_fn}
#     p(\mathbf{y}\mid \mathbf{f}) = \prod_{n=1}^N p(y_n\mid f(x_n))\,.
# \end{align}
# $$
# Conceptually, this conditional distribution describes the probability of the observed
# data, conditional on the latent function values.
#
# ### Why is the likelihood important?
#
# Choosing the correct likelihood function when building a GP, or any Bayesian model
# for that matter, is crucial. The likelihood function encodes our assumptions about
# the data and the noise that we expect to observe. For example, if we are modelling
# air pollution, then we would not expect to observe negative values of pollution. In
# this case, we would choose a likelihood function that is only defined for positive
# values. Similarly, if our data is the proportion of people who voted for a particular
# political party, then we would expect to observe values between 0 and 1. In this
# case, we would choose a likelihood function that is only defined for values between
# 0 and 1.
#
# ## Likelihoods in GPJax
#
# In GPJax, all likelihoods are a subclass of the `AbstractLikelihood` class. This base
# abstract class contains the three core methods that all likelihoods must implement:
# `predict`, `link_function`, and `expected_log_likelihood`. We will discuss each of
# these methods in the forthcoming sections, but first, we will show how to instantiate
# a likelihood object. To do this, we'll need a dataset.

# %%
import jax

# Enable Float64 for more stable matrix inversions.
from jax import config
import jax.numpy as jnp
import jax.random as jr
import matplotlib.pyplot as plt

from examples.utils import use_mpl_style
import gpjax as gpx

config.update("jax_enable_x64", True)


# set the default style for plotting
use_mpl_style()
cols = plt.rcParams["axes.prop_cycle"].by_key()["color"]

key = jr.key(42)

n = 50
x = jnp.sort(jr.uniform(key=key, shape=(n, 1), minval=-3.0, maxval=3.0), axis=0)
xtest = jnp.linspace(-3, 3, 100)[:, None]
f = lambda x: jnp.sin(x)
y = f(x) + 0.1 * jr.normal(key, shape=x.shape)
D = gpx.Dataset(x, y)

fig, ax = plt.subplots()
ax.plot(x, y, "o", label="Observations")
ax.plot(x, f(x), label="Latent function")
ax.legend()

# %% [markdown]
# In this example, our observations have support $[-3, 3]$ and are generated from a
# sinusoidal function with Gaussian noise. As such, our response values $\mathbf{y}$
# range between $-1$ and $1$, subject to Gaussian noise. Due to this, a Gaussian
# likelihood is appropriate for this dataset as it allows for negative values.
#
# As we see in \eqref{eq:likelihood_fn}, the likelihood function factorises over the
# $n$ observations. As such, we must provide this information to GPJax when
# instantiating a likelihood object. We do this by specifying the `num_datapoints`
# argument.

# %%
gpx.likelihoods.Gaussian(num_datapoints=D.n)

# %% [markdown]
# ### Likelihood parameters
#
# Some likelihoods, such as the Gaussian likelihood, contain parameters that we seek
# to infer. In the case of the Gaussian likelihood, we have a single parameter
# $\sigma^2$ that determines the observation noise. In GPJax, we can specify the value
# of $\sigma$ when instantiating the likelihood object. If we do not specify a
# value, then the likelihood will be initialised with a default value. In the case of
# the Gaussian likelihood, the default value is $1.0$. If we instead wanted to
# initialise the likelihood standard deviation with a value of $0.5$, then we would do
# this as follows:

# %%
gpx.likelihoods.Gaussian(num_datapoints=D.n, obs_stddev=0.5)

# %% [markdown]
#
# ### Prediction
#
# The `predict` method of a likelihood object transforms the latent distribution of
# the Gaussian process. In the case of a Gaussian likelihood, this simply applies the
# observational noise value to the diagonal values of the covariance matrix. For other
# likelihoods, this may be a more complex transformation. For example, the Bernoulli
# likelihood transforms the latent distribution of the Gaussian process into a
# distribution over binary values.
#
# We visualise this below for the Gaussian likelihood function. In blue we can see
# samples of $\mathbf{f}^{\star}$, whilst in red we see samples of
# $\mathbf{y}^{\star}$.

# %%
kernel = gpx.kernels.Matern32()
meanf = gpx.mean_functions.Zero()
prior = gpx.gps.Prior(kernel=kernel, mean_function=meanf)

likelihood = gpx.likelihoods.Gaussian(num_datapoints=D.n, obs_stddev=0.1)

posterior = prior * likelihood

latent_dist = posterior.predict(xtest, D)

fig, axes = plt.subplots(ncols=3, nrows=1, figsize=(9, 2))
key, subkey = jr.split(key)

for ax in axes.ravel():
    subkey, _ = jr.split(subkey)
    ax.plot(
        latent_dist.sample(sample_shape=(1,), key=subkey).T,
        lw=1,
        color=cols[0],
        label="Latent samples",
    )
    ax.plot(
        likelihood.predict(latent_dist).sample(sample_shape=(1,), key=subkey).T,
        "o",
        markersize=5,
        alpha=0.3,
        color=cols[1],
        label="Predictive samples",
    )

# %% [markdown]
# Similarly, for a Bernoulli likelihood function, the samples of $y$ would be binary.

# %%
likelihood = gpx.likelihoods.Bernoulli(num_datapoints=D.n)


fig, axes = plt.subplots(ncols=3, nrows=1, figsize=(9, 2))
key, subkey = jr.split(key)

for ax in axes.ravel():
    subkey, _ = jr.split(subkey)
    ax.plot(
        latent_dist.sample(sample_shape=(1,), key=subkey).T,
        lw=1,
        color=cols[0],
        label="Latent samples",
    )
    ax.plot(
        likelihood.predict(latent_dist).sample(sample_shape=(1,), key=subkey).T,
        "o",
        markersize=3,
        alpha=0.5,
        color=cols[1],
        label="Predictive samples",
    )

# %% [markdown]
# ### Link functions
#
# In the above figure, we can see the latent samples being constrained to be either 0 or
# 1 when a Bernoulli likelihood is specified. This is achieved by the
# `inverse link_function` $\eta(\cdot)$ of the likelihood. The link function is a
# deterministic function that maps the latent distribution of the Gaussian process to
# the support of the likelihood function. For example, the link function of the
# Bernoulli likelihood that is used in GPJax is the inverse probit function
# $$
# \eta(x) = 0.5\left(1 + \Phi\left(\frac{x}{\sqrt{2}}\right) * (1-2)\right)\,,
# $$
# where $\Phi$ is the cumulative distribution function of the standard normal
# distribution.
#
# A table of commonly used link functions and their corresponding likelihood can be
# found [here](https://en.wikipedia.org/wiki/Generalized_linear_model#Link_function).
#
# ### Expected log likelihood
#
# The final method that is associated with a likelihood function in GPJax is the
# expected log-likelihood. This term is evaluated in the
# [stochastic variational Gaussian process](uncollapsed_vi.md) in the ELBO term. For a
# variational approximation $q(f)= \mathcal{N}(f\mid m, S)$, the ELBO can be written as
#
# $$
# \begin{align}
#     \mathcal{L}(q) = \mathbb{E}_{f\sim q(f)}\left[ p(\mathbf{y}\mid f)\right] - \mathrm{KL}\left(q(f)\mid\mid p(f)\right)\,.
# \end{align}
# $$
#
# As both $q(f)$ and $p(f)$ are Gaussian distributions, the Kullback-Leibler term can
# be analytically computed. However, the expectation term is not always so easy to
# compute. Fortunately, the bound in \eqref{eq:elbo} can be decomposed as a sum of the
# datapoints
#
# $$
# \begin{align}
#     \mathcal{L}(q) = \sum_{n=1}^N \mathbb{E}_{f\sim q(f)}\left[ p(y_n\mid f)\right] - \mathrm{KL}\left(q(f)\mid\mid p(f)\right)\,.
# \end{align}
# $$
#
# This simplifies computation of the expectation as it is now a series of $N$
# 1-dimensional integrals. As such, GPJax by default uses quadrature to compute these
# integrals. However, for some likelihoods, such as the Gaussian likelihood, the
# expectation can be computed analytically. In these cases, we can supply an object
# that inherits from `AbstractIntegrator` to the likelihood upon instantiation. To see
# this, let us consider a Gaussian likelihood where we'll first define a variational
# approximation to the posterior.

# %%
z = jnp.linspace(-3.0, 3.0, 10).reshape(-1, 1)
q = gpx.variational_families.VariationalGaussian(posterior=posterior, inducing_inputs=z)


def q_moments(x):
    qx = q(x)
    return qx.mean, qx.variance


mean, variance = jax.vmap(q_moments)(x[:, None])

# %% [markdown]
# Now that we have the variational mean and variational (co)variance, we can compute
# the expected log-likelihood using the `expected_log_likelihood` method of the
# likelihood object.

# %%
jnp.sum(likelihood.expected_log_likelihood(y=y, mean=mean, variance=variance))

# %% [markdown]
# However, had we wanted to do this using quadrature, then we would have done the
# following:

# %%
lquad = gpx.likelihoods.Gaussian(
    num_datapoints=D.n,
    obs_stddev=jnp.array([0.1]),
    integrator=gpx.integrators.GHQuadratureIntegrator(num_points=20),
)

# %% [markdown]
# However, this is not recommended for the Gaussian likelihood given that the
# expectation can be computed analytically.

# %% [markdown]
# ## System configuration

# %%
# %reload_ext watermark
# %watermark -n -u -v -iv -w -a 'Thomas Pinder'



# Copyright 2022 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import (
    Callable,
    Tuple,
)

from jax import config
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import (
    Array,
    Float,
)
import numpy as np
import numpyro.distributions as npd
import pytest

from gpjax.likelihoods import (
    Bernoulli,
    Gaussian,
    Poisson,
    inv_probit,
)

# Enable Float64 for more stable matrix inversions.
config.update("jax_enable_x64", True)
_initialise_key = jr.key(123)


def _compute_latent_dist(
    n: int,
) -> Tuple[npd.MultivariateNormal, Float[Array, " N"], Float[Array, "N N"]]:
    k1, k2 = jr.split(_initialise_key)
    latent_mean = jr.uniform(k1, shape=(n,))
    latent_sqrt = jr.uniform(k2, shape=(n, n))
    latent_cov = jnp.matmul(latent_sqrt, latent_sqrt.T)
    latent_dist = npd.MultivariateNormal(loc=latent_mean, covariance_matrix=latent_cov)
    return latent_dist, latent_mean, latent_cov


@pytest.mark.parametrize("n", [1, 2, 10, 3])
@pytest.mark.parametrize("obs_stddev", [0.1, 0.5, 1.0, 0.0])
def test_gaussian_likelihood(n: int, obs_stddev: float):
    x = jnp.linspace(-3.0, 3.0).reshape(-1, 1)
    likelihood = Gaussian(num_datapoints=n, obs_stddev=obs_stddev)

    assert isinstance(likelihood.link_function, Callable)
    assert isinstance(likelihood.link_function(x), npd.Normal)

    # Construct latent function distribution.
    latent_dist, latent_mean, latent_cov = _compute_latent_dist(n)
    pred_dist = likelihood(latent_dist)
    assert isinstance(pred_dist, npd.MultivariateNormal)

    # Check predictive mean and variance.
    assert (pred_dist.mean == latent_mean).all()
    noise_matrix = jnp.eye(likelihood.num_datapoints) * likelihood.obs_stddev.value**2
    assert np.allclose(
        pred_dist.scale_tril, jnp.linalg.cholesky(latent_cov + noise_matrix)
    )


@pytest.mark.parametrize("n", [1, 2, 10])
def test_bernoulli_likelihood(n: int):
    x = jnp.linspace(-3.0, 3.0).reshape(-1, 1)
    likelihood = Bernoulli(num_datapoints=n)

    assert isinstance(likelihood.link_function, Callable)
    assert isinstance(likelihood.link_function(x), npd.BernoulliProbs)

    # Construct latent function distribution.
    latent_dist, latent_mean, latent_cov = _compute_latent_dist(n)
    pred_dist = likelihood(latent_dist)
    assert isinstance(pred_dist, npd.BernoulliProbs)

    # Check predictive mean and variance.
    p = inv_probit(latent_mean / jnp.sqrt(1.0 + jnp.diagonal(latent_cov)))
    assert (pred_dist.mean == p).all()
    assert (pred_dist.variance == p * (1.0 - p)).all()


@pytest.mark.parametrize("n", [1, 2, 10])
def test_poisson_likelihood(n: int):
    x = jnp.linspace(-3.0, 3.0).reshape(-1, 1)
    likelihood = Poisson(num_datapoints=n)

    assert isinstance(likelihood.link_function, Callable)
    assert isinstance(likelihood.link_function(x), npd.Poisson)

    # Construct latent function distribution.
    latent_dist, latent_mean, latent_cov = _compute_latent_dist(n)
    pred_dist = likelihood(latent_dist)
    assert isinstance(pred_dist, npd.Poisson)

    # Check predictive mean and variance.
    rate = jnp.exp(latent_mean)
    assert (pred_dist.mean == rate).all()



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.11.2
#   kernelspec:
#     display_name: docs
#     language: python
#     name: python3
# ---

# %%
from dataclasses import (
    dataclass,
    field,
)

# %%
from beartype.typing import (
    Any,
    Callable,
    Dict,
)
import jax.numpy as jnp  # noqa: F401
import jupytext

# %%
import gpjax

# %%
get_last = lambda x: x[-1]


# %%
@dataclass
class Result:
    path: str
    comparisons: field(default_factory=dict)  # type: ignore
    precision: int = 1
    compare_history: bool = True

    def __post_init__(self):
        self.name: str = self.path.split("/")[-1].split(".")[0].replace("_", "-")

    def _compare(
        self,
        observed_variables: Dict[str, Any],
        variable_name: str,
        true_value: float,
        operation: Callable[[Any], Any],
    ):
        if variable_name == "history" and not self.compare_history:
            return
        try:
            value = operation(observed_variables[variable_name])
            assert abs(true_value - value) < self.precision
        except AssertionError as e:
            print(e)

    def test(self):
        notebook = jupytext.read(self.path)
        contents = ""
        for c in notebook["cells"]:
            if c["cell_type"] == "code":
                if c["source"].startswith("%"):
                    pass
                else:
                    contents += c["source"]
            contents += "\n"

        contents = contents.replace('plt.style.use("./gpjax.mplstyle")', "").replace(
            "plt.show()", ""
        )
        lines = contents.split("\n")
        contents = "\n".join([line for line in lines if not line.startswith("%")])

        loc = {}

        # weird bug in interactive interpreter: lambda functions
        # don't have access to the global scope of the executed file
        # so we need to pass gpjax in the globals explicitly
        # since it's used in a lambda function inside the examples
        _globals = globals()
        _globals["gpx"] = gpjax
        exec(contents, _globals, loc)
        for k, v in self.comparisons.items():
            truth, op = v
            self._compare(
                observed_variables=loc, variable_name=k, true_value=truth, operation=op
            )


# %%
regression = Result(
    path="examples/regression.py",
    comparisons={
        "history": (55.07405622, get_last),
        "predictive_mean": (36.24383416, jnp.sum),
        "predictive_std": (197.04727051, jnp.sum),
    },
)
regression.test()

# %%
sparse = Result(
    path="examples/collapsed_vi.py",
    comparisons={
        "history": (1924.7634809, get_last),
        "predictive_mean": (-8.39869652, jnp.sum),
        "predictive_std": (255.74838027, jnp.sum),
    },
)
sparse.test()

# %%
stochastic = Result(
    path="examples/uncollapsed_vi.py",
    comparisons={
        "history": (-2678.41302494, get_last),
        "meanf": (-54.14787028, jnp.sum),
        "sigma": (121.4298333, jnp.sum),
    },
)
stochastic.test()



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import beartype.typing as tp
import jax.numpy as jnp
from jaxtyping import (
    Float,
    Int,
    Num,
)

from gpjax.kernels.computations import (
    AbstractKernelComputation,
    EigenKernelComputation,
)
from gpjax.kernels.non_euclidean.utils import jax_gather_nd
from gpjax.kernels.stationary.base import StationaryKernel
from gpjax.parameters import (
    Parameter,
    PositiveReal,
    Static,
)
from gpjax.typing import (
    Array,
    ScalarFloat,
    ScalarInt,
)


class GraphKernel(StationaryKernel):
    r"""The Matérn graph kernel defined on the vertex set of a graph.

    A Matérn graph kernel defined on the vertices of a graph.

    Computes the covariance for pairs of vertices $(v_i, v_j)$ with variance $\sigma^2$:
    $$
    k(v_i, v_j) = \sigma^2 \exp\Bigg(-\frac{\lVert v_i - v_j \rVert^2_2}{2\ell^2}\Bigg)
    $$
    where $\ell$ is the lengthscale parameter and $\sigma^2$ is the variance.

    The key reference for this object is Borovitskiy et. al., (2020).

    """

    num_vertex: tp.Union[ScalarInt, None]
    laplacian: Static[Float[Array, "N N"]]
    eigenvalues: Static[Float[Array, "N 1"]]
    eigenvectors: Static[Float[Array, "N N"]]
    name: str = "Graph Matérn"

    def __init__(
        self,
        laplacian: Num[Array, "N N"],
        active_dims: tp.Union[list[int], slice, None] = None,
        lengthscale: tp.Union[ScalarFloat, Float[Array, " D"], Parameter] = 1.0,
        variance: tp.Union[ScalarFloat, Parameter] = 1.0,
        smoothness: ScalarFloat = 1.0,
        n_dims: tp.Union[int, None] = None,
        compute_engine: AbstractKernelComputation = EigenKernelComputation(),
    ):
        """Initializes the kernel.

        Args:
            laplacian: the Laplacian matrix of the graph.
            active_dims: The indices of the input dimensions that the kernel operates on.
            lengthscale: the lengthscale(s) of the kernel ℓ. If a scalar or an array of
                length 1, the kernel is isotropic, meaning that the same lengthscale is
                used for all input dimensions. If an array with length > 1, the kernel is
                anisotropic, meaning that a different lengthscale is used for each input.
            variance: the variance of the kernel σ.
            smoothness: the smoothness parameter of the Matérn kernel.
            n_dims: The number of input dimensions. If `lengthscale` is an array, this
                argument is ignored.
            compute_engine: The computation engine that the kernel uses to compute the
                covariance matrix.
        """
        if isinstance(smoothness, Parameter):
            self.smoothness = smoothness
        else:
            self.smoothness = PositiveReal(smoothness)

        self.laplacian = Static(laplacian)
        evals, eigenvectors = jnp.linalg.eigh(self.laplacian.value)
        self.eigenvectors = Static(eigenvectors)
        self.eigenvalues = Static(evals.reshape(-1, 1))
        self.num_vertex = self.eigenvalues.value.shape[0]

        super().__init__(active_dims, lengthscale, variance, n_dims, compute_engine)

    def __call__(  # TODO not consistent with general kernel interface
        self,
        x: Int[Array, "N 1"],
        y: Int[Array, "N 1"],
        *,
        S,
        **kwargs,
    ):
        Kxx = (jax_gather_nd(self.eigenvectors.value, x) * S.squeeze()) @ jnp.transpose(
            jax_gather_nd(self.eigenvectors.value, y)
        )  # shape (n,n)
        return Kxx.squeeze()



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.7
#   kernelspec:
#     display_name: gpjax
#     language: python
#     name: python3
# ---

# %% [markdown]
# # UCI Data Benchmarking
#
# In this notebook, we will show how to apply GPJax on a benchmark UCI regression
# problem. These kind of tasks are often used in the research community to benchmark
# and assess new techniques against those already in the literature. Much of the code
# contained in this notebook can be adapted to applied problems concerning datasets
# other than the one presented here.

# %%
# Enable Float64 for more stable matrix inversions.
from jax import config
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import install_import_hook
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.metrics import (
    mean_squared_error,
    r2_score,
)
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler

from examples.utils import use_mpl_style

config.update("jax_enable_x64", True)


with install_import_hook("gpjax", "beartype.beartype"):
    import gpjax as gpx


# set the default style for plotting
use_mpl_style()
cols = mpl.rcParams["axes.prop_cycle"].by_key()["color"]

key = jr.key(42)

# %% [markdown]
# ## Data Loading
#
# We'll be using the
# [Yacht](https://archive.ics.uci.edu/ml/datasets/yacht+hydrodynamics) dataset from
# the UCI machine learning data repository. Each observation describes the
# hydrodynamic performance of a yacht through its resistance. The dataset contains 6
# covariates and a single positive, real valued response variable. There are 308
# observations in the dataset, so we can comfortably use a conjugate regression
# Gaussian process here (for more more details, checkout the
# [Regression notebook](https://docs.jaxgaussianprocesses.com/_examples/regression/)).

# %%
try:
    yacht = pd.read_fwf("data/yacht_hydrodynamics.data", header=None).values[:-1, :]
except FileNotFoundError:
    yacht = pd.read_fwf(
        "docs/_examples/data/yacht_hydrodynamics.data", header=None
    ).values[:-1, :]

X = yacht[:, :-1]
y = yacht[:, -1].reshape(-1, 1)

# %% [markdown]
# ## Preprocessing
#
# With a dataset loaded, we'll now preprocess it such that it is more amenable to
# modelling with a Gaussian process.
#
# ### Data Partitioning
#
# We'll first partition our data into a _training_ and _testing_ split. We'll fit our
# Gaussian process to the training data and evaluate its performance on the test data.
# This allows us to investigate how effectively our Gaussian process generalises to
# out-of-sample datapoints and ensure that we are not overfitting. We'll hold 30% of
# our data back for testing purposes.

# %%
Xtr, Xte, ytr, yte = train_test_split(X, y, test_size=0.3, random_state=42)

# %% [markdown]
# ### Response Variable
#
# We'll now process our response variable $\mathbf{y}$. As the below plots show, the
# data has a very long tail and is certainly not Gaussian. However, we would like to
# model a Gaussian response variable so that we can adopt a Gaussian likelihood
# function and leverage the model's conjugacy. To achieve this, we'll first log-scale
# the data, to bring the long right tail in closer to the data's mean. We'll then
# standardise the data such that is distributed according to a unit normal
# distribution. Both of these transformations are invertible through the log-normal
# expectation and variance formulae and the the inverse standardisation identity,
# should we ever need our model's predictions to be back on the scale of the
# original dataset.
#
# For transforming both the input and response variable, all transformations will be
# done with respect to the training data where relevant.

# %%
log_ytr = np.log(ytr)
log_yte = np.log(yte)

y_scaler = StandardScaler().fit(log_ytr)
scaled_ytr = y_scaler.transform(log_ytr)
scaled_yte = y_scaler.transform(log_yte)

# %% [markdown]
# We can see the effect of these transformations in the below three panels.

# %%
fig, ax = plt.subplots(ncols=3, figsize=(9, 2.5))
ax[0].hist(ytr, bins=30, color=cols[1])
ax[0].set_title("y")
ax[1].hist(log_ytr, bins=30, color=cols[1])
ax[1].set_title("log(y)")
ax[2].hist(scaled_ytr, bins=30, color=cols[1])
ax[2].set_title("scaled log(y)")

# %% [markdown]
# ### Input Variable
#
# We'll now transform our input variable $\mathbf{X}$ to be distributed according to a
# unit Gaussian.

# %%
x_scaler = StandardScaler().fit(Xtr)
scaled_Xtr = x_scaler.transform(Xtr)
scaled_Xte = x_scaler.transform(Xte)

# %% [markdown]
# ## Model fitting
#
# With data now loaded and preprocessed, we'll proceed to defining a Gaussian process
# model and optimising its parameters. This notebook purposefully does not go into
# great detail on this process, so please see notebooks such as the
# [Regression notebook](https://docs.jaxgaussianprocesses.com/_examples/regression/)
# and
# [Classification notebook](https://docs.jaxgaussianprocesses.com/_examples/classification)
# for further information.
#
# ### Model specification
#
# We'll use a radial basis function kernel to parameterise the Gaussian process in this
# notebook. As we have 5 covariates, we'll assign each covariate its own lengthscale
# parameter. This form of kernel is commonly known as an automatic relevance
# determination (ARD) kernel.
#
# In practice, the exact form of kernel used should be selected such that it
# represents your understanding of the data. For example, if you were to model
# temperature; a process that we know to be periodic, then you would likely wish to
# select a periodic kernel. Having _Gaussian-ised_ our data somewhat, we'll also adopt
# a Gaussian likelihood function.

# %%
n_train, n_covariates = scaled_Xtr.shape
kernel = gpx.kernels.RBF(
    active_dims=list(range(n_covariates)),
    variance=jnp.var(scaled_ytr),
    lengthscale=0.1 * jnp.ones((n_covariates,)),
)
meanf = gpx.mean_functions.Zero()
prior = gpx.gps.Prior(mean_function=meanf, kernel=kernel)

likelihood = gpx.likelihoods.Gaussian(num_datapoints=n_train)

posterior = prior * likelihood

# %% [markdown]
# ### Model Optimisation
#
# With a model now defined, we can proceed to optimise the hyperparameters of our
# model using Scipy.

# %%
training_data = gpx.Dataset(X=scaled_Xtr, y=scaled_ytr)

opt_posterior, history = gpx.fit_scipy(
    model=posterior,
    # we use the negative mll as we are minimising
    objective=lambda p, d: -gpx.objectives.conjugate_mll(p, d),
    train_data=training_data,
)

print(-gpx.objectives.conjugate_mll(opt_posterior, training_data))

# %% [markdown]
# ## Prediction
#
# With an optimal set of parameters learned, we can make predictions on the set of
# data that we held back right at the start. We'll do this in the usual way by first
# computing the latent function's distribution before computing the predictive
# posterior distribution.

# %%
latent_dist = opt_posterior(scaled_Xte, training_data)
predictive_dist = likelihood(latent_dist)

predictive_mean = predictive_dist.mean
predictive_stddev = jnp.sqrt(predictive_dist.variance)

# %% [markdown]
# ## Evaluation
#
# We'll now show how the performance of our Gaussian process can be evaluated by
# numerically and visually.
#
# ### Metrics
#
# To numerically assess the performance of our model, two commonly used metrics are
# root mean squared error (RMSE) and the R2 coefficient. RMSE is simply the square
# root of the squared difference between predictions and actuals. A value of 0 for
# this metric implies that our model has 0 generalisation error on the test set. R2
# measures the amount of variation within the data that is explained by the model.
# This can be useful when designing variance reduction methods such as control
# variates as it allows you to understand what proportion of the data's variance will
# be soaked up. A perfect model here would score 1 for R2 score, whereas predicting
# the data's mean would score 0 and models doing worse than simple mean predictions
# can score less than 0.

# %%
rmse = mean_squared_error(y_true=scaled_yte.squeeze(), y_pred=predictive_mean)
r2 = r2_score(y_true=scaled_yte.squeeze(), y_pred=predictive_mean)
print(f"Results:\n\tRMSE: {rmse: .4f}\n\tR2: {r2: .2f}")

# %% [markdown]
# Both of these metrics seem very promising, so, based off these, we can be quite
# happy that our first attempt at modelling the Yacht data is promising.
#
# ### Diagnostic plots
#
# To accompany the above metrics, we can also produce residual plots to explore
# exactly where our model's shortcomings lie. If we define a residual as the true
# value minus the prediction, then we can produce three plots:
#
# 1. Predictions vs. actuals.
# 2. Predictions vs. residuals.
# 3. Residual density.
#
# The first plot allows us to explore if our model struggles to predict well for
# larger or smaller values by observing where the model deviates more from the line
# $y=x$. In the second plot we can inspect whether or not there were outliers or
# structure within the errors of our model. A well-performing model would have
# predictions close to and symmetrically distributed either side of $y=0$. Such a
# plot can be useful for diagnosing heteroscedasticity. Finally, by plotting a
# histogram of our residuals we can observe whether or not there is any skew to
# our residuals.

# %%
residuals = scaled_yte.squeeze() - predictive_mean

fig, ax = plt.subplots(ncols=3, figsize=(9, 2.5), tight_layout=True)

ax[0].scatter(predictive_mean, scaled_yte.squeeze(), color=cols[1])
ax[0].plot([0, 1], [0, 1], color=cols[0], transform=ax[0].transAxes)
ax[0].set(xlabel="Predicted", ylabel="Actual", title="Predicted vs Actual")

ax[1].scatter(predictive_mean.squeeze(), residuals, color=cols[1])
ax[1].plot([0, 1], [0.5, 0.5], color=cols[0], transform=ax[1].transAxes)
ax[1].set_ylim([-1.0, 1.0])
ax[1].set(xlabel="Predicted", ylabel="Residuals", title="Predicted vs Residuals")

ax[2].hist(np.asarray(residuals), bins=30, color=cols[1])
ax[2].set_title("Residuals")

# %% [markdown]
# From this, we can see that our model is struggling to predict the smallest values
# of the Yacht's hydrodynamic and performs increasingly well as the Yacht's
# hydrodynamic performance increases. This is likely due to the original data's heavy
# right-skew, and successive modelling attempts may wish to introduce a
# heteroscedastic likelihood function that would enable more flexible modelling of
# the smaller response values.
#
# ## System configuration

# %%
# %reload_ext watermark
# %watermark -n -u -v -iv -w -a 'Thomas Pinder'



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from itertools import product
from typing import Any

from cola.ops.operator_base import LinearOperator
import jax
from jax import config
import jax.numpy as jnp
import jax.random as jr
import pytest

from gpjax.kernels.base import AbstractKernel
from gpjax.kernels.computations import AbstractKernelComputation
from gpjax.kernels.nonstationary import (
    ArcCosine,
    Linear,
    Polynomial,
)
from gpjax.parameters import (
    NonNegativeReal,
    Static,
)

# Enable Float64 for more stable matrix inversions.
config.update("jax_enable_x64", True)


def params_product(params: dict[str, list]) -> list[dict[str, Any]]:
    return [
        dict(zip(params.keys(), values, strict=False))
        for values in product(*params.values())
    ]


TESTED_KERNELS = [
    (
        ArcCosine,
        params_product(
            {
                "order": [0, 1, 2],
                "weight_variance": [0.1, 1.0],
                "bias_variance": [0.1, 1.0],
            }
        ),
    ),
    (Linear, [{}]),
    (Polynomial, params_product({"degree": [1, 2, 3], "shift": [1e-6, 0.1, 1.0]})),
]

VARIANCES = [0.1]


@pytest.fixture
def kernel_request(
    kernel,
    params,
    variance,
):
    return kernel, params, variance


@pytest.mark.parametrize(
    "kernel, params", [(cls, p) for cls, params in TESTED_KERNELS for p in params]
)
@pytest.mark.parametrize("variance", VARIANCES)
@pytest.fixture
def test_init(kernel_request):
    kernel, params, variance = kernel_request
    return kernel(**params, variance=variance)


@pytest.mark.parametrize(
    "kernel, params", [(cls, p) for cls, params in TESTED_KERNELS for p in params]
)
@pytest.mark.parametrize("variance", VARIANCES)
def test_init_override_paramtype(kernel_request):
    kernel, params, variance = kernel_request

    new_params = {}  # otherwise we change the fixture and next test fails
    for param, value in params.items():
        if param in ("degree", "order"):
            continue
        new_params[param] = Static(value)

    k = kernel(**new_params, variance=NonNegativeReal(variance))
    assert isinstance(k.variance, NonNegativeReal)

    for param in params.keys():
        if param in ("degree", "order"):
            continue
        assert isinstance(getattr(k, param), Static)


@pytest.mark.parametrize("kernel", [k[0] for k in TESTED_KERNELS])
def test_init_defaults(kernel: type[AbstractKernel]):
    # Initialise kernel
    k = kernel()

    # Check that the parameters are set correctly
    assert isinstance(k.compute_engine, type(AbstractKernelComputation()))
    assert isinstance(k.variance, NonNegativeReal)


@pytest.mark.parametrize("kernel", [k[0] for k in TESTED_KERNELS])
@pytest.mark.parametrize("variance", VARIANCES)
def test_init_variances(kernel: type[AbstractKernel], variance):
    # Initialise kernel
    k = kernel(variance=variance)

    # Check that the parameters are set correctly
    assert isinstance(k.variance, NonNegativeReal)
    assert jnp.allclose(k.variance.value, jnp.asarray(variance))

    # Check that error is raised if variance is not valid
    with pytest.raises(ValueError):
        k = kernel(variance=-1.0)

    with pytest.raises(TypeError):
        k = kernel(variance=jnp.ones((2, 2)))

    with pytest.raises(TypeError):
        k = kernel(variance="invalid type")


@pytest.mark.parametrize(
    "kernel, params", [(cls, p) for cls, params in TESTED_KERNELS for p in params]
)
@pytest.mark.parametrize("variance", VARIANCES)
@pytest.mark.parametrize("n", [1, 2, 5], ids=lambda x: f"n={x}")
def test_gram(test_init: AbstractKernel, n: int):
    # kernel is initialized in the test_init fixture
    k = test_init
    n_dims = k.n_dims or 1

    # Inputs
    x = jnp.linspace(0.0, 1.0, n * n_dims).reshape(n, n_dims)

    # Test gram matrix
    Kxx = k.gram(x)
    assert isinstance(Kxx, LinearOperator)
    assert Kxx.shape == (n, n)
    assert jnp.all(jnp.linalg.eigvalsh(Kxx.to_dense() + jnp.eye(n) * 1e-6) > 0.0)


@pytest.mark.parametrize(
    "kernel, params", [(cls, p) for cls, params in TESTED_KERNELS for p in params]
)
@pytest.mark.parametrize("variance", VARIANCES)
@pytest.mark.parametrize("n_a", [1, 2, 5], ids=lambda x: f"n_a={x}")
@pytest.mark.parametrize("n_b", [1, 2, 5], ids=lambda x: f"n_b={x}")
def test_cross_covariance(test_init: AbstractKernel, n_a: int, n_b: int):
    # kernel is initialized in the test_init fixture
    k = test_init
    n_dims = k.n_dims or 1

    # Inputs
    x = jnp.linspace(0.0, 1.0, n_a * n_dims).reshape(n_a, n_dims)
    y = jnp.linspace(0.0, 1.0, n_b * n_dims).reshape(n_b, n_dims)

    # Test cross covariance matrix
    Kxy = k.cross_covariance(x, y)
    assert isinstance(Kxy, jax.Array)
    assert Kxy.shape == (n_a, n_b)


@pytest.mark.parametrize("order", [0, 1, 2])
def test_arccosine_special_case(order: int):
    """For certain values of weight variance (1.0) and bias variance (0.0), we can test
    our calculations using the Monte Carlo expansion of the arccosine kernel, e.g.
    see Eq. (1) of https://cseweb.ucsd.edu/~saul/papers/nips09_kernel.pdf.
    """
    kernel = ArcCosine(
        weight_variance=jnp.array([1.0, 1.0]), bias_variance=1e-25, order=order
    )

    # Inputs close(ish) together
    a = jnp.array([[0.0, 0.0]])
    b = jnp.array([[2.0, 2.0]])

    # calc cross-covariance exactly
    Kab_exact = kernel.cross_covariance(a, b)

    # calc cross-covariance using samples
    weights = jax.random.normal(jr.PRNGKey(123), (10_000, 2))  # [S, d]
    weights_a = jnp.matmul(weights, a.T)  # [S, 1]
    weights_b = jnp.matmul(weights, b.T)  # [S, 1]
    H_a = jnp.heaviside(weights_a, 0.5)
    H_b = jnp.heaviside(weights_b, 0.5)
    integrands = H_a * H_b * (weights_a**order) * (weights_b**order)
    Kab_approx = 2.0 * jnp.mean(integrands)

    assert jnp.max(Kab_approx - Kab_exact) < 1e-4



# Copyright 2022 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import abc
import functools as ft

import beartype.typing as tp
from flax import nnx
import jax.numpy as jnp
from jaxtyping import (
    Float,
    Num,
)

from gpjax.parameters import (
    Parameter,
    Real,
    Static,
)
from gpjax.typing import (
    Array,
    ScalarFloat,
)


class AbstractMeanFunction(nnx.Module):
    r"""Mean function that is used to parameterise the Gaussian process."""

    def __init__(self) -> None:
        super().__init__()

    @abc.abstractmethod
    def __call__(self, x: Num[Array, "N D"]) -> Float[Array, "N O"]:
        r"""Evaluate the mean function at the given points. This method is required for all subclasses.

        Args:
            x (Float[Array, " D"]): The point at which to evaluate the mean function.

        Returns:
            Float[Array, "1]: The evaluated mean function.
        """
        raise NotImplementedError

    def __add__(
        self, other: tp.Union["AbstractMeanFunction", Float[Array, " O"]]
    ) -> "AbstractMeanFunction":
        r"""Add two mean functions.

        Args:
            other (AbstractMeanFunction): The other mean function to add.

        Returns:
            AbstractMeanFunction: The sum of the two mean functions.
        """
        if isinstance(other, AbstractMeanFunction):
            return SumMeanFunction([self, other])

        return SumMeanFunction([self, Constant(other)])

    def __radd__(
        self,
        other: tp.Union[
            "AbstractMeanFunction", Float[Array, " O"]
        ],  # TODO should this be ScalarFloat? or Num?
    ) -> "AbstractMeanFunction":
        r"""Add two mean functions.

        Args:
            other (AbstractMeanFunction): The other mean function to add.

        Returns:
            AbstractMeanFunction: The sum of the two mean functions.
        """
        return self.__add__(other)

    def __mul__(
        self,
        other: tp.Union[
            "AbstractMeanFunction", Float[Array, " O"]
        ],  # TODO should this be ScalarFloat? or Num?
    ) -> "AbstractMeanFunction":
        r"""Multiply two mean functions.

        Args:
            other (AbstractMeanFunction): The other mean function to multiply.

        Returns:
            AbstractMeanFunction: The product of the two mean functions.
        """
        if isinstance(other, AbstractMeanFunction):
            return ProductMeanFunction([self, other])

        return ProductMeanFunction([self, Constant(other)])

    def __rmul__(
        self,
        other: tp.Union[
            "AbstractMeanFunction", Float[Array, " O"]
        ],  # TODO should this be ScalarFloat? or Num?
    ) -> "AbstractMeanFunction":
        r"""Multiply two mean functions.

        Args:
            other (AbstractMeanFunction): The other mean function to multiply.

        Returns:
            AbstractMeanFunction: The product of the two mean functions.
        """
        return self.__mul__(other)


class Constant(AbstractMeanFunction):
    r"""Constant mean function.

    A constant mean function. This function returns a repeated scalar value for all
    inputs.  The scalar value itself can be treated as a model hyperparameter and
    learned during training but defaults to 1.0.
    """

    def __init__(
        self,
        constant: tp.Union[ScalarFloat, Float[Array, " O"], Parameter, Static] = 0.0,
    ):
        if isinstance(constant, Parameter) or isinstance(constant, Static):
            self.constant = constant
        else:
            self.constant = Real(jnp.array(constant))

    def __call__(self, x: Num[Array, "N D"]) -> Float[Array, "N O"]:
        r"""Evaluate the mean function at the given points.

        Args:
            x (Float[Array, " D"]): The point at which to evaluate the mean function.

        Returns:
            Float[Array, "1"]: The evaluated mean function.
        """
        return jnp.ones((x.shape[0], 1)) * self.constant.value


class Zero(Constant):
    r"""Zero mean function.

    The zero mean function. This function returns a zero scalar value for all
    inputs. Unlike the Constant mean function, the constant scalar zero is fixed, and
    cannot be treated as a model hyperparameter and learned during training.
    """

    def __init__(self):
        super().__init__(constant=Static(jnp.array(0.0)))


class CombinationMeanFunction(AbstractMeanFunction):
    r"""A base class for products or sums of AbstractMeanFunctions."""

    def __init__(
        self,
        means: list[AbstractMeanFunction],
        operator: tp.Callable,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        # Add means to a list, flattening out instances of this class therein, as in GPFlow kernels.
        items_list: list[AbstractMeanFunction] = []

        for item in means:
            if not isinstance(item, AbstractMeanFunction):
                raise TypeError(
                    "can only combine AbstractMeanFunction instances"
                )  # pragma: no cover

            if isinstance(item, self.__class__):
                items_list.extend(item.means)
            else:
                items_list.append(item)

        self.means = items_list
        self.operator = operator

    def __call__(self, x: Num[Array, "N D"]) -> Float[Array, "N O"]:
        r"""Evaluate combination kernel on a pair of inputs.

        Args:
            x (Float[Array, " D"]): The point at which to evaluate the mean function.

        Returns:
            Float[Array, " Q"]: The evaluated mean function.
        """
        return self.operator(jnp.stack([m(x) for m in self.means]))


SumMeanFunction = ft.partial(
    CombinationMeanFunction, operator=ft.partial(jnp.sum, axis=0)
)
ProductMeanFunction = ft.partial(
    CombinationMeanFunction, operator=ft.partial(jnp.prod, axis=0)
)



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import jax.numpy as jnp
from jaxtyping import Float
import numpyro.distributions as npd

from gpjax.kernels.stationary.base import StationaryKernel
from gpjax.kernels.stationary.utils import squared_distance
from gpjax.typing import (
    Array,
    ScalarFloat,
)


class RBF(StationaryKernel):
    r"""The Radial Basis Function (RBF) kernel.

    Computes the covariance for pair of inputs $(x, y)$ with lengthscale parameter
    $\ell$ and variance $\sigma^2$:
    $$
    k(x,y)=\sigma^2\exp\Bigg(- \frac{\lVert x - y \rVert^2_2}{2 \ell^2} \Bigg)
    $$
    """

    name: str = "RBF"

    def __call__(self, x: Float[Array, " D"], y: Float[Array, " D"]) -> ScalarFloat:
        x = self.slice_input(x) / self.lengthscale.value
        y = self.slice_input(y) / self.lengthscale.value
        K = self.variance.value * jnp.exp(-0.5 * squared_distance(x, y))
        return K.squeeze()

    @property
    def spectral_density(self) -> npd.Normal:
        return npd.Normal(0.0, 1.0)



from pathlib import Path

import mkdocs_gen_files
from mkdocs_gen_files import Nav
from typing import Iterable, Mapping, Union
import dataclasses


class CustomNav(Nav):
    def __init__(self):
        super().__init__()

    @dataclasses.dataclass
    class Item:
        level: int
        title: str
        filename: Union[str, None]

    @classmethod
    def _items(cls, data: Mapping, level: int) -> Iterable[Item]:
        for key, value in data.items():
            if key is not None:
                if key == "gpjax":
                    title = "GPJax"
                elif key == "gps":
                    title = "GPs"
                elif key == "rbf":
                    title = "RBF"
                elif key == "rff":
                    title = "RFF"
                else:
                    title = key.title()

                title = title.replace("_", " ")
                title = title.replace("Matern", "Matérn")

                yield cls.Item(level=level, title=title, filename=value.get(None))
                yield from cls._items(value, level + 1)


nav = CustomNav()

for path in sorted(Path("gpjax").rglob("*.py")):
    module_path = path.relative_to(".").with_suffix("")
    doc_path = path.relative_to("gpjax").with_suffix(".md")
    full_doc_path = Path("api", doc_path)

    parts = list(module_path.parts)

    if parts[-1] == "__init__":
        parts = parts[:-1]
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
        continue
    elif parts[-1] == "__main__":
        continue

    # final_part = parts[-1].title()
    # parts = parts[:-1] + [final_part]
    nav[parts] = doc_path.as_posix()

    # print(full_doc_path)
    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        identifier = ".".join(parts)
        title = identifier.split(".")[-1].replace("_", " ").title()
        if title == "Gps":
            title = "GPs"
        elif title == "Rbf":
            title = "RBF"
        elif title == "Rff":
            title = "RFF"

        if "Matern" in title:
            title = title.replace("Matern", "Matérn")

        # print(f"# {title}\n", file=fd)
        print("::: " + identifier, file=fd)  #

    mkdocs_gen_files.set_edit_path(full_doc_path, path)

with mkdocs_gen_files.open("api/SUMMARY.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import beartype.typing as tp
from flax import nnx
import jax.numpy as jnp
from jaxtyping import Float
import numpyro.distributions as npd

from gpjax.kernels.base import AbstractKernel
from gpjax.kernels.computations import (
    AbstractKernelComputation,
    DenseKernelComputation,
)
from gpjax.parameters import (
    NonNegativeReal,
    PositiveReal,
)
from gpjax.typing import (
    Array,
    ScalarArray,
    ScalarFloat,
)

Lengthscale = tp.Union[Float[Array, "D"], ScalarArray]
LengthscaleCompatible = tp.Union[ScalarFloat, list[float], Lengthscale]


class StationaryKernel(AbstractKernel):
    """Base class for stationary kernels.

    Stationary kernels are a class of kernels that are invariant to translations
    in the input space. They can be isotropic or anisotropic, meaning that they
    can have a single lengthscale for all input dimensions or a different lengthscale
    for each input dimension.
    """

    lengthscale: nnx.Variable[Lengthscale]
    variance: nnx.Variable[ScalarArray]

    def __init__(
        self,
        active_dims: tp.Union[list[int], slice, None] = None,
        lengthscale: tp.Union[LengthscaleCompatible, nnx.Variable[Lengthscale]] = 1.0,
        variance: tp.Union[ScalarFloat, nnx.Variable[ScalarArray]] = 1.0,
        n_dims: tp.Union[int, None] = None,
        compute_engine: AbstractKernelComputation = DenseKernelComputation(),
    ):
        """Initializes the kernel.

        Args:
            active_dims: The indices of the input dimensions that the kernel operates on.
            lengthscale: the lengthscale(s) of the kernel ℓ. If a scalar or an array of
                length 1, the kernel is isotropic, meaning that the same lengthscale is
                used for all input dimensions. If an array with length > 1, the kernel is
                anisotropic, meaning that a different lengthscale is used for each input.
            variance: the variance of the kernel σ.
            n_dims: The number of input dimensions. If `lengthscale` is an array, this
                argument is ignored.
            compute_engine: The computation engine that the kernel uses to compute the
                covariance matrix.
        """

        super().__init__(active_dims, n_dims, compute_engine)
        self.n_dims = _validate_lengthscale(lengthscale, self.n_dims)
        if isinstance(lengthscale, nnx.Variable):
            self.lengthscale = lengthscale
        else:
            self.lengthscale = PositiveReal(lengthscale)

            # static typing
            if tp.TYPE_CHECKING:
                self.lengthscale = tp.cast(PositiveReal[Lengthscale], self.lengthscale)

        if isinstance(variance, nnx.Variable):
            self.variance = variance
        else:
            self.variance = NonNegativeReal(variance)

            # static typing
            if tp.TYPE_CHECKING:
                self.variance = tp.cast(NonNegativeReal[ScalarFloat], self.variance)

    @property
    def spectral_density(self) -> npd.Normal | npd.StudentT:
        r"""The spectral density of the kernel.

        Returns:
            Callable[[Float[Array, "D"]], Float[Array, "D"]]: The spectral density function.
        """
        raise NotImplementedError(
            f"Kernel {self.name} does not have a spectral density."
        )


def _validate_lengthscale(
    lengthscale: tp.Union[LengthscaleCompatible, nnx.Variable[Lengthscale]],
    n_dims: tp.Union[int, None],
):
    # Check that the lengthscale is a valid value.
    _check_lengthscale(lengthscale)

    n_dims = _check_lengthscale_dims_compat(lengthscale, n_dims)
    return n_dims


def _check_lengthscale_dims_compat(
    lengthscale: tp.Union[LengthscaleCompatible, nnx.Variable[Lengthscale]],
    n_dims: tp.Union[int, None],
):
    r"""Check that the lengthscale is compatible with n_dims.

    If possible, infer the number of input dimensions from the lengthscale.
    """

    if isinstance(lengthscale, nnx.Variable):
        return _check_lengthscale_dims_compat_old(lengthscale.value, n_dims)

    lengthscale = jnp.asarray(lengthscale)
    ls_shape = jnp.shape(lengthscale)

    if ls_shape == ():
        return n_dims
    elif ls_shape != () and n_dims is None:
        return ls_shape[0]
    elif ls_shape != () and n_dims is not None:
        if ls_shape != (n_dims,):
            raise ValueError(
                "Expected `lengthscale` to be compatible with the number "
                f"of input dimensions. Got `lengthscale` with shape {ls_shape}, "
                f"but the number of input dimensions is {n_dims}."
            )
        return n_dims


def _check_lengthscale_dims_compat_old(
    lengthscale: tp.Union[LengthscaleCompatible, nnx.Variable[Lengthscale]],
    n_dims: tp.Union[int, None],
):
    r"""Check that the lengthscale is compatible with n_dims.

    If possible, infer the number of input dimensions from the lengthscale.
    """

    if isinstance(lengthscale, nnx.Variable):
        return _check_lengthscale_dims_compat_old(lengthscale.value, n_dims)

    lengthscale = jnp.asarray(lengthscale)
    ls_shape = jnp.shape(lengthscale)

    if ls_shape == ():
        return lengthscale, n_dims
    elif ls_shape != () and n_dims is None:
        return lengthscale, ls_shape[0]
    elif ls_shape != () and n_dims is not None:
        if ls_shape != (n_dims,):
            raise ValueError(
                "Expected `lengthscale` to be compatible with the number "
                f"of input dimensions. Got `lengthscale` with shape {ls_shape}, "
                f"but the number of input dimensions is {n_dims}."
            )
        return lengthscale, n_dims


def _check_lengthscale(lengthscale: tp.Any):
    """Check that the lengthscale is a valid value."""

    if isinstance(lengthscale, nnx.Variable):
        _check_lengthscale(lengthscale.value)
        return

    if not isinstance(lengthscale, (int, float, jnp.ndarray, list, tuple)):
        raise TypeError(
            f"Expected `lengthscale` to be a array-like. Got {lengthscale}."
        )

    if isinstance(lengthscale, (jnp.ndarray, list)):
        ls_shape = jnp.shape(jnp.asarray(lengthscale))

        if len(ls_shape) > 1:
            raise ValueError(
                f"Expected `lengthscale` to be a scalar or 1D array. "
                f"Got `lengthscale` with shape {ls_shape}."
            )



from jax import config

config.update("jax_enable_x64", True)

import jax.numpy as jnp
import pytest

from gpjax.citation import (
    AbstractCitation,
    NullCitation,
    PaperCitation,
    PhDThesisCitation,
    cite,
)
from gpjax.kernels import (
    RBF,
    RFF,
    ArcCosine,
    GraphKernel,
    Linear,
    Matern12,
    Matern32,
    Matern52,
)


def _check_no_fallback(citation: AbstractCitation):
    # Check the fallback has not been used
    assert repr(citation) != repr(
        "No citation available. If you think this is an error, please open a pull"
        " request."
    )
    assert (
        str(citation)
        != "No citation available. If you think this is an error, please open a pull"
        " request."
    )


@pytest.mark.parametrize("kernel", [Matern12(), Matern32(), Matern52()])
def test_matern_kernels(kernel):
    citation = cite(kernel)
    # Check type
    assert isinstance(citation, PhDThesisCitation)
    # Check some fields
    assert citation.authors == "Bertil Matérn"
    assert citation.year == "1960"
    assert citation.institution == "Stockholm University"

    _check_no_fallback(citation)


def test_arc_cosine():
    kernel = ArcCosine()
    citation = cite(kernel)

    assert isinstance(citation, PaperCitation)

    assert citation.citation_key == "cho2009kernel"
    assert citation.title == "Kernel Methods for Deep Learning"
    assert citation.authors == "Cho, Youngmin and Saul, Lawrence"
    _check_no_fallback(citation)


def test_graph_kernel():
    L = jnp.eye(3)
    kernel = GraphKernel(
        laplacian=L,
    )
    citation = cite(kernel)

    assert isinstance(citation, PaperCitation)
    assert citation.citation_key == "borovitskiy2021matern"
    assert citation.title == "Matérn Gaussian Processes on Graphs"
    assert citation.year == "2021"
    _check_no_fallback(citation)


@pytest.mark.parametrize("kernel", [RBF, Matern12, Matern32, Matern52])
def test_rff(kernel):
    base_kernel = kernel(n_dims=1)
    rff = RFF(base_kernel=base_kernel)
    citation = cite(rff)

    assert isinstance(citation, PaperCitation)
    assert citation.citation_key == "rahimi2007random"
    assert citation.title == "Random features for large-scale kernel machines"
    assert citation.authors == "Rahimi, Ali and Recht, Benjamin"
    assert citation.year == "2007"
    _check_no_fallback(citation)


@pytest.mark.parametrize("kernel", [RBF(), Linear()])
def test_missing_citation(kernel):
    citation = cite(kernel)
    assert isinstance(citation, NullCitation)



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.7
#   kernelspec:
#     display_name: gpjax
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Deep Kernel Learning
#
# In this notebook we demonstrate how GPJax can be used in conjunction with
# [Flax](https://flax.readthedocs.io/en/latest/) to build deep kernel Gaussian
# processes. Modelling data with discontinuities is a challenging task for regular
# Gaussian process models. However, as shown in
# <strong data-cite="wilson2016deep"></strong>, transforming the inputs to our
# Gaussian process model's kernel through a neural network can offer a solution to this.

# %%
from dataclasses import (
    dataclass,
    field,
)

from flax import nnx
import jax

# Enable Float64 for more stable matrix inversions.
from jax import config
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import (
    Array,
    Float,
    install_import_hook,
)
import matplotlib as mpl
import matplotlib.pyplot as plt
import optax as ox
from scipy.signal import sawtooth

from examples.utils import use_mpl_style
from gpjax.kernels.computations import (
    AbstractKernelComputation,
    DenseKernelComputation,
)

config.update("jax_enable_x64", True)


with install_import_hook("gpjax", "beartype.beartype"):
    import gpjax as gpx
    from gpjax.kernels.base import AbstractKernel


# set the default style for plotting
use_mpl_style()
cols = mpl.rcParams["axes.prop_cycle"].by_key()["color"]

key = jr.key(42)


# %% [markdown]
# ## Dataset
#
# As previously mentioned, deep kernels are particularly useful when the data has
# discontinuities. To highlight this, we will use a sawtooth function as our data.

# %%
n = 500
noise = 0.2

key, subkey = jr.split(key)
x = jr.uniform(key=key, minval=-2.0, maxval=2.0, shape=(n,)).reshape(-1, 1)
f = lambda x: jnp.asarray(sawtooth(2 * jnp.pi * x))
signal = f(x)
y = signal + jr.normal(subkey, shape=signal.shape) * noise

D = gpx.Dataset(X=x, y=y)

xtest = jnp.linspace(-2.0, 2.0, 500).reshape(-1, 1)
ytest = f(xtest)

fig, ax = plt.subplots()
ax.plot(x, y, "o", label="Training data", alpha=0.5)
ax.plot(xtest, ytest, label="True function")
ax.legend(loc="best")

# %% [markdown]
# ## Deep kernels
#
# ### Details
#
# Instead of applying a kernel $k(\cdot, \cdot')$ directly on some data, we seek to
# apply a _feature map_ $\phi(\cdot)$ that projects the data to learn more meaningful
# representations beforehand. In deep kernel learning, $\phi$ is a neural network
# whose parameters are learned jointly with the GP model's hyperparameters. The
# corresponding kernel is then computed by $k(\phi(\cdot), \phi(\cdot'))$. Here
# $k(\cdot,\cdot')$ is referred to as the _base kernel_.
#
# ### Implementation
#
# Although deep kernels are not currently supported natively in GPJax, defining one is
# straightforward as we now demonstrate. Inheriting from the base `AbstractKernel`
# in GPJax, we create the `DeepKernelFunction` object that allows the
# user to supply the neural network and base kernel of their choice. Kernel matrices
# are then computed using the regular `gram` and `cross_covariance` functions.


# %%
@dataclass
class DeepKernelFunction(AbstractKernel):
    base_kernel: AbstractKernel
    network: nnx.Module
    compute_engine: AbstractKernelComputation = field(
        default_factory=lambda: DenseKernelComputation()
    )

    def __call__(
        self, x: Float[Array, " D"], y: Float[Array, " D"]
    ) -> Float[Array, "1"]:
        xt = self.network(x)
        yt = self.network(y)
        return self.base_kernel(xt, yt)


# %% [markdown]
# ### Defining a network
#
# With a deep kernel object created, we proceed to define a neural network. Here we
# consider a small multi-layer perceptron with two linear hidden layers and ReLU
# activation functions between the layers. The first hidden layer contains 64 units,
# while the second layer contains 32 units. Finally, we'll make the output of our
# network a three units wide. The corresponding kernel that we define will then be of
# [ARD form](https://docs.jaxgaussianprocesses.com/_examples/constructing_new_kernels/#active-dimensions)
# to allow for different lengthscales in each dimension of the feature space.
# Users may wish to design more intricate network structures for more complex tasks,
# which functionality is supported well in Haiku.


# %%
feature_space_dim = 3


class Network(nnx.Module):
    def __init__(
        self, rngs: nnx.Rngs, *, input_dim: int, inner_dim: int, feature_space_dim: int
    ) -> None:
        self.layer1 = nnx.Linear(input_dim, inner_dim, rngs=rngs)
        self.output_layer = nnx.Linear(inner_dim, feature_space_dim, rngs=rngs)
        self.rngs = rngs

    def __call__(self, x: jax.Array) -> jax.Array:
        x = x.reshape((x.shape[0], -1))
        x = self.layer1(x)
        x = jax.nn.relu(x)
        x = self.output_layer(x).squeeze()
        return x


forward_linear = Network(
    nnx.Rngs(123), feature_space_dim=feature_space_dim, inner_dim=32, input_dim=1
)

# %% [markdown]
# ## Defining a model
#
# Having characterised the feature extraction network, we move to define a Gaussian
# process parameterised by this deep kernel. We consider a third-order Matérn base
# kernel and assume a Gaussian likelihood.

# %%
base_kernel = gpx.kernels.Matern52(
    active_dims=list(range(feature_space_dim)),
    lengthscale=jnp.ones((feature_space_dim,)),
)
kernel = DeepKernelFunction(network=forward_linear, base_kernel=base_kernel)
meanf = gpx.mean_functions.Zero()
prior = gpx.gps.Prior(mean_function=meanf, kernel=kernel)
likelihood = gpx.likelihoods.Gaussian(num_datapoints=D.n)
posterior = prior * likelihood
# %% [markdown]
# ### Optimisation
#
# We train our model via maximum likelihood estimation of the marginal log-likelihood.
# The parameters of our neural network are learned jointly with the model's
# hyperparameter set.
#
# With the inclusion of a neural network, we take this opportunity to highlight the
# additional benefits gleaned from using
# [Optax](https://optax.readthedocs.io/en/latest/) for optimisation. In particular, we
# showcase the ability to use a learning rate scheduler that decays the optimiser's
# learning rate throughout the inference. We decrease the learning rate according to a
# half-cosine curve over 700 iterations, providing us with large step sizes early in
# the optimisation procedure before approaching more conservative values, ensuring we
# do not step too far. We also consider a linear warmup, where the learning rate is
# increased from 0 to 1 over 50 steps to get a reasonable initial learning rate value.

# %%
schedule = ox.warmup_cosine_decay_schedule(
    init_value=0.0,
    peak_value=0.01,
    warmup_steps=75,
    decay_steps=700,
    end_value=0.0,
)

optimiser = ox.chain(
    ox.clip(1.0),
    ox.adamw(learning_rate=schedule),
)

opt_posterior, history = gpx.fit(
    model=posterior,
    objective=lambda p, d: -gpx.objectives.conjugate_mll(p, d),
    train_data=D,
    optim=optimiser,
    num_iters=800,
    key=key,
)

# %% [markdown]
# ## Prediction
#
# With a set of learned parameters, the only remaining task is to predict the output
# of the model. We can do this by simply applying the model to a test data set.

# %%
latent_dist = opt_posterior(xtest, train_data=D)
predictive_dist = opt_posterior.likelihood(latent_dist)

predictive_mean = predictive_dist.mean
predictive_std = jnp.sqrt(predictive_dist.variance)

fig, ax = plt.subplots()
ax.plot(x, y, "o", label="Observations", color=cols[0])
ax.plot(xtest, predictive_mean, label="Predictive mean", color=cols[1])
ax.fill_between(
    xtest.squeeze(),
    predictive_mean - 2 * predictive_std,
    predictive_mean + 2 * predictive_std,
    alpha=0.2,
    color=cols[1],
    label="Two sigma",
)
ax.plot(
    xtest,
    predictive_mean - 2 * predictive_std,
    color=cols[1],
    linestyle="--",
    linewidth=1,
)
ax.plot(
    xtest,
    predictive_mean + 2 * predictive_std,
    color=cols[1],
    linestyle="--",
    linewidth=1,
)
ax.legend()

# %% [markdown]
# ## System configuration

# %%
# %reload_ext watermark
# %watermark -n -u -v -iv -w -a 'Thomas Pinder'
